#*******************************************************************************
# E.S.O. - VLT project
#
# "@(#) $Id: mxu_class.tcl,v 2.43 2004/06/17 23:04:33 vltsccm Exp $"
#
# who       when        what
# --------  ----------  --------------------------------------------------------
# tsz/pbg   2003/11/13  clarified warning in method M2W
# aaguayo   2003/10/28  PPRS 10127 - warning about ref-slit out of range
# aaguayo   2000/12/12  added ref-slit label for CURVED slit type 
#                       (to display warning later) 
# W.Hummel  02.MAR.99   adopted from ref-targ class 
# mkiekebu 2004-06-14    Fixed problem with dialog windows (RH9.0)
#
#*******************************************************************************
#   NAME             
#   MXU_SLIT - Tcl class           
# 
#   NAMESPACE
#       fors
#
#   SYNOPSIS         
#       MXU_SLIT <instance_name> canvas image ?-eqx <equinox>?
#       
#   PARENT CLASS
#       none 
#
#   DESCRIPTION      
#       Handling of a single MXU slit
#
#   OPTIONS
#
#   PUBLIC METHODS
#       plot_m {} 
#          plot current slit or circle instance to canvas
#       plot_label {oriA oriB} 
#          plots labels of current slit instance to canvas 
#       erase {} 
#          delete current slit instance from canvas
#       put_OC {a} 
#          a is slit center as a list of world coords, 
#          corrected for optical distortion 
#          set slit center in world coords and focplane coords 
#       get_OC_RA {}
#          return slit center, corrected for optical distortion 
#       get_OC_DEC {}
#          return slit center, corrected for optical distortion 
#       get_RA {}
#          return slit center, not corrected for optical distortion 
#       get_DEC {}
#          return slit center, not corrected for optical distortion 
#       get_xy {}
#          returns slit center as list in mm foc field 
#       get_slit {}
#          returns slit position and dimension in mm foc field 
#       get_oc_slit {}
#          returns slit position and dimension in mm foc field 
#          corrected for optical distortion 
#       get_oc_slit_contour {} 
#          return the 4 corner points of rectangular slits 
#          in mm foc field 
#       get_xy_mmu {}
#          returns slit center as list in mm mmu working table
#       get_slit_mmu {}
#          returns slit position and dimension in mm mmu working table
#       get_oc_slit_mmu {}
#          returns slit position and dimension in mm mmu working table
#          corrected for optical distortion 
#       get_slit_contour_mmu {} 
#          return the 4 corner points of rectangular slits 
#          in mm mmu working table NOT corrected for OC
#       get_oc_slit_contour_mmu {} 
#          return the 4 corner points of rectangular slits 
#          in mm mmu working table 
#
#   PUBLIC DATA MEMBERS
#       eqx         2000.0
#
#   COMMON METHODS
#       nos {}
#           returns the number of created slits 
#           the deleted slits are included 
#       fc {}
#           set the focal scale         
#
#   PRIVATE METHODS
#       create_slit {}
#           set the s-matrix containing slit dimensions in mm focplane 
#       create_circle {}
#           set the s-matrix containing circle dimensions in mm focplane 
#
#   PRIVATE DATA MEMBERS
#       m
#
#   COMMANDS
#       MXU_SLIT auto {m_image m_canvas equinox} 
#
#   RETURN VALUES
#
#   CAUTIONS
#       Use FIMS V1.3.1 only with skycat V2.2  
#
#   EXAMPLES
#
#   SEE ALSO
#
#-------------------------------------------------------------------------------
#
#23456789 123456789 123456789 123456789 123456789 123456789 123456789 1234567890
#


itk::usual FORS {}

# create a class for the application

namespace eval fors {

class MXU_SLIT {

inherit CONV_MC

#------------------------
constructor { m_c m_i xc yc args } {
#------------------------
     # puts stdout "mxu_class constructor for #[expr $slit_counter +1]"

     set m_image  $m_i;             # set private variable 
     set m_canvas $m_c;             # set private variable 

     eval configure $args

     if {$mmff} {
     # input coords are read from FAD, units =  mmff
     # a fixed mask, no oc corrections 
     set x $xc
     set y $yc 
     set dummy [m2c $x $y]
     set xc [lindex $dummy 0]
     set yc [lindex $dummy 1]

     set oc_x 0.0; set oc_y 0.0; set oc_alp 0.0; set oc_del 0.0

     # puts stdout "mmff $x $y -> canvas $xc $yc"

     
     } else {

     # (xc,yc) from canvas coords to (x,y) in mm focalfield
     set dummy [c2m $xc $yc]  
     set x [lindex $dummy 0]
     set y [lindex $dummy 1]
     # puts stdout "mmff $x $y <- canvas $xc $yc"
     }

     incr slit_counter 1; # common counter

     switch -exact -- [$this cget -shape] {
         STRAIGHT { create_slit; plot_m
                    if {$label} { plot_label e w }
                  }
         CURVED   { # puts stdout "constructor: create_curve"
                    # -------------------
                    # default slit shape 
                    # -------------------
                    # in mm_ff plane units 

                    set x1 [expr $x-$sl_mm]
                    set y1 [expr $y-$sl_mm/2]
                    set x2 $x
                    set y2 $y
                    set x3 [expr $x-$sl_mm]
                    set y3 [expr $y+$sl_mm/2]

                    create_curve; plot_m
                    if {$label} { plot_label e w }
                    set mc_off [ c2m 0.0 0.0 ]

                    set bf1 [list 0 0]; set bf2 [list 0 0] 
                    set bf3 [list 0 0]; set bf4 [list 0 0] 
                    set bf5 [list 0 0]; set bf6 [list 0 0]

                  }
         CIRCLE   { create_circle;  plot_m
                    if {$label} { plot_label e w }
                  }
         none     { 
                    puts stdout "shape='none' not supported" 
                    $this configure -ON 0
                    return 
                    # do nothing } 
         default  { 
                    puts stdout "mxu slit shape error" 
                    $this configure -ON 0
                    return
                  }
            }


     # set the private variable alp del in wcs 
     # for all three slit types :x,y xs,ys alp,del are the 
     # the coords of the curser click or center/gauss 
     # for curved slits this point is initially not on the slit 
     # and will not stay there if the curved slit is modified

     catch {$m_image convert coords $xc $yc canvas alp del "wcs $eqx"}
        

     set slit_initialized 1; # flag to avoid multiple init
     }

     #-----------
     destructor {
     #-----------
     incr slit_counter -1 
     }





# 1 -------------- create 


     #------------------------------
     private method create_slit {} {
     #------------------------------
     # >>>> all coords in mmff 
     # calculates contours of slit (a,b,c,...) again 
     # from slit dimensions (width, length)
     set xin [lindex $ff 0];  # focal plane border, XSRMIN or XHRMIN 
     set xax [lindex $ff 1];  # focal plane border, XSRMAX or XHRMAX

     # focal field coordinates
     # using an order separation filter OG_.. GG_..
     # one slit per dispersion stripe 
     #
     # posang == 0 !!
     #
     #                NORTH
     #  |                             |
     #              b   c              
     # a+ - - - - - +---+ - - - - - - +d
     #              |   |              
     #  |           |   |             |      WEST
     #              |   |              
     # e+ - - - - - +---+ - - - - - - +h
     #              f   g             
     #  |                             |
     #               SOUTH 
     #
     # posang > 0 !!
     #                NORTH
     #  |                             |
     #              b   c              
     # a+ - - - - - +---+ - - - - - - +d
     #               \   \             
     #  |             \   \           |      WEST
     #                 \   \           
     # e+ - - - - - - - +---+ - - - - +h
     #                  f   g             
     #  |                             |
     #               SOUTH 
     #
     # using interference filter (e.g. H_ALPHA+35)
     # more than 1 slit per interference stripe 
     # posang == 0 !!
     #
     #                NORTH
     #  |                               |
     #          a     b   c     d        
     #  +   -   + - - +---+ - - +   -   + 
     #          |     |   |     |        
     #  |       |     |   |     |       |      WEST
     #          |     |   |     |        
     #  +   -   + - - +---+ - - +   -   + 
     #          e     f   g     h       
     #  |                               |
     #               SOUTH 
     #
     # posang > 0 !!
     #                NORTH
     #  |                             |
     #        a     b   c     d        
     #  + - - + - - +---+ - - + - - - +
     #         \     \   \     \       
     #  |       \     \   \     \     |      WEST
     #           \     \   \     \     
     #  + - - - - + - - +---+ - - + - +
     #            e     f   g     h       
     #  |                             |
     #               SOUTH 

     # puts stdout "create slit : $x $y"

     # solution for order separation or broad band filters 
     set s(xa) $xin; set s(ya) [expr $y - $sl_mm*0.5] 
     set s(xd) $xax; set s(yd) $s(ya)
     set s(xb) [expr $x+(-$sw_mm-$sl_mm*tan($pa))*0.5]; set s(yb) $s(ya) 
     set s(xc) [expr $x+($sw_mm-$sl_mm*tan($pa))*0.5]; set s(yc) $s(ya) 

     set s(xe) $xin; set s(ye) [expr $y+ $sl_mm*0.5] 
     set s(xh) $xax; set s(yh) $s(ye)
     set s(xf) [expr $x+(-$sw_mm+$sl_mm*tan($pa))*0.5]; set s(yf) $s(ye) 
     set s(xg) [expr $x+($sw_mm+$sl_mm*tan($pa))*0.5]; set s(yg) $s(ye) 

     # solution for interference filters 
     if {[string length $fwhm]>0 && $fwhm < 600.0} {
      set s(xa) [expr $x-$sl_mm*tan($pa)*0.5+($lcf-$lcg-$fwhm)*$disp]; 
                set s(ya) [expr $y- $sl_mm*0.5] 
      set s(xd) [expr $x-$sl_mm*tan($pa)*0.5+($lcf-$lcg+$fwhm)*$disp]; 
                 set s(yd) $s(ya)
      set s(xb) [expr $x+(-$sw_mm-$sl_mm*tan($pa))*0.5]; set s(yb) $s(ya) 
      set s(xc) [expr $x+($sw_mm-$sl_mm*tan($pa))*0.5]; set s(yc) $s(ya) 

      set s(xe) [expr $x+$sl_mm*tan($pa)*0.5+($lcf-$lcg-$fwhm)*$disp]; 
                set s(ye) [expr $y+ $sl_mm*0.5] 
      set s(xh) [expr $x+$sl_mm*tan($pa)*0.5+($lcf-$lcg+$fwhm)*$disp]; 
                set s(yh) $s(ye)
      set s(xf) [expr $x+(-$sw_mm+$sl_mm*tan($pa))*0.5]; set s(yf) $s(ye) 
      set s(xg) [expr $x+($sw_mm+$sl_mm*tan($pa))*0.5]; set s(yg) $s(ye) 
      }


     if {$pa==0} {
         # derive slit number from instance name 
         set a [lindex [split $this : ] 6]  
         set i [string trimleft $a C_mxus]

         # define aperture definition string for postscript 
         set add "%"
         append add "ADD"
         append add [ format "%1d" [expr $i+10] ]
         append add "R,"
         append add [format "%4.3f" $sl_mm]
         append add "X"
         append add [format "%4.3f" $sw_mm]
         append add "*%"

         # define aperture definition string for MMU   
         set add_mmu "%"
         append add_mmu "ADD"
         append add_mmu [ format "%1d" [expr $i+10] ]
         append add_mmu "R,"
         append add_mmu [format "%4.3f" $sl_mm]
         append add_mmu "X"
         append add_mmu [format "%4.3f" $sw_mm]
         append add_mmu "*%"
        } else {
        set add ""
        set add_mmu ""
        }

     # from mm focal plane (s-array) to 
     # canvas coordinates (m-array)
     foreach i { a b c d e f g h} {
           set dummy [m2c $s(x$i) $s(y$i)]
           set m(x$i) [lindex $dummy 0]
           set m(y$i) [lindex $dummy 1]
           }
     }

     #------------------------------
     public method update_slit {} {
     #------------------------------
     # from mm focal plane (s-array) to 
     # canvas coordinates (m-array)
     foreach i { a b c d e f g h} {
           set dummy [m2c $s(x$i) $s(y$i)]
           set m(x$i) [lindex $dummy 0]
           set m(y$i) [lindex $dummy 1]
           }
     }
      

     #------------------------------
     private method create_circle {} {
     #------------------------------
     # >>>> all coords in mmff 
     # calculates contours of slit (a,b,c,...) again 
     # from slit dimensions (width, length)
     set xin [lindex $ff 0];  # focal plane border, XSRMIN or XHRMIN 
     set xax [lindex $ff 1];  # focal plane border, XSRMAX or XHRMAX

     # focal field coordinates
     # using an order separation filter OG_.. GG_..
     # one slit per dispersion stripe 
     #
     # posang == 0 !!
     #
     #                NORTH
     #  |                             |
     #              b   c              
     # a+ - - - - - +-----+ - - - - - - +d
     #              |/   \|              
     #  |           |     |             |      WEST
     #              |\   /|              
     # e+ - - - - - +-----+ - - - - - - +h
     #              f     g             
     #  |                               |
     #               SOUTH 
     #
     #
     # using interference filter (e.g. H_ALPHA+35)
     # more than 1 slit per interference stripe 
     # posang == 0 !!
     #
     #                NORTH
     #  |                               |
     #          a     b   c     d        
     #  +   -   + - - +-----+ - - +   -   + 
     #          |     |/   \|     |        
     #  |       |     |     |     |       |      WEST
     #          |     |\   /|     |        
     #  +   -   + - - +-----+ - - +   -   + 
     #          e     f     g     h       
     #  |                                 |
     #               SOUTH 
     #

     # we use the slit width as circle diameter
     #                                 ========
     # solution for order separation or broad band filters 
     set s(xa) $xin; set s(ya) [expr $y - $sw_mm*0.5] 
     set s(xd) $xax; set s(yd) $s(ya)
     set s(xb) [expr $x-$sw_mm*0.5]; set s(yb) $s(ya) 
     set s(xc) [expr $x+$sw_mm*0.5]; set s(yc) $s(ya) 

     set s(xe) $xin; set s(ye) [expr $y + $sw_mm*0.5] 
     set s(xh) $xax; set s(yh) $s(ye)
     set s(xf) [expr $x-$sw_mm*0.5]; set s(yf) $s(ye) 
     set s(xg) [expr $x+$sw_mm*0.5]; set s(yg) $s(ye) 

     # solution for interference filters and low dispersion grism
     if {[string length $fwhm]>0 && $fwhm < 600.0} {
      set s(xa) [expr $x+($lcf-$lcg-$fwhm)*$disp]; 
                set s(ya) [expr $y- $sw_mm*0.5] 
      set s(xd) [expr $x+($lcf-$lcg+$fwhm)*$disp]; 
                 set s(yd) $s(ya)
      set s(xb) [expr $x-$sw_mm*0.5]; set s(yb) $s(ya) 
      set s(xc) [expr $x+$sw_mm*0.5]; set s(yc) $s(ya) 

      set s(xe) [expr $x+($lcf-$lcg-$fwhm)*$disp]; 
                set s(ye) [expr $y+ $sw_mm*0.5] 
      set s(xh) [expr $x+($lcf-$lcg+$fwhm)*$disp]; 
                set s(yh) $s(ye)
      set s(xf) [expr $x-$sw_mm*0.5]; set s(yf) $s(ye) 
      set s(xg) [expr $x+$sw_mm*0.5]; set s(yg) $s(ye) 
      }


         # derive slit number from instance name 
         set a [lindex [split $this : ] 6]  
         set i [string trimleft $a C_mxus]

         # define aperture definition string for gerber 
         set add "%"
         append add "ADD"
         append add [ format "%1d" [expr $i+10] ]
         append add "C,"
         append add [format "%4.3f" $sw_mm]
         append add "*%"

         # define aperture definition string for postscript file
         set add_mmu "%"
         append add_mmu "ADD"
         append add_mmu [ format "%1d" [expr $i+10] ]
         append add_mmu "C,"
         append add_mmu [format "%4.3f" $sw_mm]
         append add_mmu "*%"

     # from mm focal plane (s-array) to 
     # canvas coordinates (m-array)
     foreach i { a b c d e f g h} {
           set dummy [m2c $s(x$i) $s(y$i)]
           set m(x$i) [lindex $dummy 0]
           set m(y$i) [lindex $dummy 1]
           }
     }

    #-------------------------------
    public method update_circle {} {
    #-------------------------------
     # from mm focal plane (s-array) to 
     # canvas coordinates (m-array)
     foreach i { a b c d e f g h} {
           set dummy [m2c $s(x$i) $s(y$i)]
           set m(x$i) [lindex $dummy 0]
           set m(y$i) [lindex $dummy 1]
           }
     }
    

     #------------------------------
     private method create_curve {} {
     #------------------------------
     # >>>> all coords in mmff 
     # calculates contours of curved slit (a,b,c,...) again 
     # this is called either by 'config slit_width' 
     # or by config bez-shape  

     # from slit dimensions (width, length)
     set xin [lindex $ff 0];  # focal plane border, XSRMIN or XHRMIN 
     set xax [lindex $ff 1];  # focal plane border, XSRMAX or XHRMAX

     # the three Bezier points are given by 
     # (b+c)/2, (p+q)/2, (f+g)/2 
     #
     # focal field coordinates
     # using an order separation filter OG_.. GG_..
     # one slit per dispersion stripe 
     #
     #                NORTH
     #
     #                (x1,y1)
     #  |             |               |
     #              b | c              
     # a+ - - - - - +-+-+ - - - - - - +d
     #               \   \              
     #                \   \              
     #               p|   |q  +(x2,y2)            
     #                /   /              
     #               /   /              
     # e+ - - - - - +-+-+ - - - - - - +h
     #              f | g             
     #  |             |               |
     #                (x3,y3)         
     #               SOUTH 
     #
     # Bezier defined by three points (x1,y1), (x,y) and (x3,y3)
     # central curve defined by bezier points
     # blue slit rim defined by b p f
     # red slit rim defined by c q g
     #
     # using interference filter (e.g. H_ALPHA+35)
     # more than 1 slit per interference stripe 
     # posang == 0 !!
     #
     #                NORTH
     #  |                               |
     #          a     b   c     d        
     #  +   -   + - - +---+ - - +   -   + 
     #          |      \   \    |        
     #  |       |      |   |    |       |      WEST
     #          |      /   /    |        
     #  +   -   + - - +---+ - - +   -   + 
     #          e     f   g     h       
     #  |                               |
     #               SOUTH 
     #

     # units: mmff


     # solution for order separation or broad band filters 
     # a,d,e,h at the border

     set s(x1) $x1; set s(y1) $y1
     set s(x2) $x2; set s(y2) $y2
     set s(x3) $x3; set s(y3) $y3

     set s(xa) $xin; set s(ya) $s(y1) 
     set s(xd) $xax; set s(yd) $s(y1)
     set s(xb) [expr $s(x1)-$sw_mm*0.5]; set s(yb) $s(y1) 
     set s(xc) [expr $s(x1)+$sw_mm*0.5]; set s(yc) $s(y1) 

     set s(xe) $xin; set s(ye) $s(y3) 
     set s(xh) $xax; set s(yh) $s(y3)
     set s(xf) [expr $s(x3)-$sw_mm*0.5]; set s(yf) $s(y3) 
     set s(xg) [expr $s(x3)+$sw_mm*0.5]; set s(yg) $s(y3) 
   
     set s(xp) [expr $s(x2)-$sw_mm*0.5]; set s(yp) $s(y2)
     set s(xq) [expr $s(x2)+$sw_mm*0.5]; set s(yq) $s(y2)


     # solution for interference filters  and low dispersion grism
     if {[string length $fwhm]>0 && $fwhm < 600.0} {
     set s(x1) $x1; set s(y1) $y1
     set s(x2) $x2; set s(y2) $y2
     set s(x3) $x3; set s(y3) $y3

      set s(xa) [expr $s(x1)+($lcf-$lcg-$fwhm)*$disp]; 
                set s(ya) $s(y1) 
      set s(xd) [expr $s(x1)+($lcf-$lcg+$fwhm)*$disp]; 
                 set s(yd) $s(y1)
      set s(xb) [expr $s(x1)-$sw_mm*0.5]; set s(yb) $s(y1) 
      set s(xc) [expr $s(x1)+$sw_mm*0.5]; set s(yc) $s(y1) 

      set s(xe) [expr $s(x3)+($lcf-$lcg-$fwhm)*$disp]; 
                set s(ye) $s(y3) 
      set s(xh) [expr $s(x3)+($lcf-$lcg+$fwhm)*$disp]; 
                set s(yh) $s(y3)
      set s(xf) [expr $s(x3)-$sw_mm*0.5]; set s(yf) $s(y3)
      set s(xg) [expr $s(x3)+$sw_mm*0.5]; set s(yg) $s(y3)

      set s(xp) [expr $s(x2)-$sw_mm*0.5]; set s(yp) $s(y2)
      set s(xq) [expr $s(x2)+$sw_mm*0.5]; set s(yq) $s(y2)
      }
        set add ""
        set add_mmu ""



     # from mm focal plane (s-array) to 
     # canvas coordinates (m-array)
     foreach i { a b c d e f g h 1 2 3 p q} {
           set dummy [m2c $s(x$i) $s(y$i)]
           set m(x$i) [lindex $dummy 0]
           set m(y$i) [lindex $dummy 1]
           }

     }


     # -----------------------------
     public method update_curve {} {
     # -----------------------------
     # from mm focal plane (s-array) to 
     # canvas coordinates (m-array)
     foreach i { a b c d e f g h 1 2 3 p q} {
           set dummy [m2c $s(x$i) $s(y$i)]
           set m(x$i) [lindex $dummy 0]
           set m(y$i) [lindex $dummy 1]
           }
     }

     #--------------------------------------
     public method update_curved_center {} {
     #--------------------------------------
     if {[string match *CURVED* [$this cget -shape]]} {

     # define the central point on the slit 
     # for saving sequence 
     # central slit point
     # ------------------
     set xc [expr 0.25*$m(x1) + 0.5*$m(x2) + 0.25*$m(x3) ] 
     set yc [expr 0.25*$m(y1) + 0.5*$m(y2) + 0.25*$m(y3) ] 
     $m_image convert coords $xc $yc canvas alp del "wcs $eqx"

     set x [expr 0.25*$s(x1) + 0.5*$s(x2) + 0.25*$s(x3) ] 
     set y [expr 0.25*$s(y1) + 0.5*$s(y2) + 0.25*$s(y3) ] 

     return [list $xc $yc] 
     }
     }


# 2 ------------------- plot 



     #----------------------
     public method plot_m {} {
     #----------------------
     # plot current MXU slit instance to canvas 

     # derive slit number from instance name 
     set a [lindex [split $this : ] 6]  
     set i [string trimleft $a C_mxus]

     # delete from canvas 
     $m_canvas delete mxu$i

     # redraw 

      switch -exact -- [$this cget -shape] {

        STRAIGHT {

     if {[string length $fwhm]>0 && $fwhm < 600.0} {
        # solution for interference filters 
        $m_canvas create poly $m(xa) $m(ya) $m(xd) $m(yd) \
           $m(xh) $m(yh) $m(xe) $m(ye) -outline $c3 \
           -fill $c1 -stipple $trapa -tag " A mm mxu$i ma$i "
        } else {
        # solution for order separation filters 
        $m_canvas create poly $m(xa) $m(ya) $m(xb) $m(yb) \
           $m(xf) $m(yf) $m(xe) $m(ye) -outline $c3 \
           -fill $c1 -stipple $trapa -tag " A mm mxu$i ma$i "

        $m_canvas create poly $m(xd) $m(yd) $m(xc) $m(yc) \
           $m(xg) $m(yg) $m(xh) $m(yh) -outline $c3 \
           -fill $c1 -stipple $trapa -tag " A mm mxu$i mb$i "
        }

     $m_canvas create poly $m(xb) $m(yb) $m(xc) $m(yc) \
        $m(xg) $m(yg) $m(xf) $m(yf) -outline $c3 \
        -fill "" -stipple "" -tag " A mm mxu$i ms$i "
            }

      CIRCLE {

     if {[string length $fwhm]>0 && $fwhm < 600.0} {
        # solution for interference filters 
        $m_canvas create poly $m(xa) $m(ya) $m(xd) $m(yd) \
           $m(xh) $m(yh) $m(xe) $m(ye) -outline $c3 \
           -fill $c1 -stipple $trapa -tag " A mm mxu$i ma$i "
        } else {
        # solution for order separation filters 
        $m_canvas create poly $m(xa) $m(ya) $m(xb) $m(yb) \
           $m(xf) $m(yf) $m(xe) $m(ye) -outline $c3 \
           -fill $c1 -stipple $trapa -tag " A mm mxu$i ma$i "

        $m_canvas create poly $m(xd) $m(yd) $m(xc) $m(yc) \
           $m(xg) $m(yg) $m(xh) $m(yh) -outline $c3 \
           -fill $c1 -stipple $trapa -tag " A mm mxu$i mb$i "
        }

     $m_canvas create poly $m(xb) $m(yb) $m(xc) $m(yc) \
        $m(xg) $m(yg) $m(xf) $m(yf) -outline $c3 \
        -fill "" -stipple "" -tag " A mm mxu$i ms$i "
      
                  # tcltk too stupid, cannot draw oval with angle
                  set lx [expr $m(xb)-$m(xg)]; set ly [expr $m(yb)-$m(yg)]
                  set lr [expr sqrt($lx*$lx+$ly*$ly)/2.0/sqrt(2.0)]
                  set lx [expr ($m(xb) +$m(xg)) / 2.0]
                  set ly [expr ($m(yb) +$m(yg)) / 2.0]
     
                  $m_canvas create oval [expr $lx-$lr] [expr $ly-$lr] \
                                        [expr $lx+$lr] [expr $ly+$lr] \
                     -outline $c3 -fill "" -tag " A mm mxu$i ms$i "
              }
       
      CURVED {


     if {[string length $fwhm]>0 && $fwhm < 600.0} {
        # solution for interference filters 
        $m_canvas create poly $m(xa) $m(ya) $m(xd) $m(yd) \
           $m(xh) $m(yh) $m(xe) $m(ye) -outline $c3 \
           -fill $c1 -stipple $trapa -tag " A mm mxu$i ma$i "
        } else {

        # solution for order separation filters 
        $m_canvas create line $m(xb) $m(yb) $m(xa) $m(ya) \
           $m(xe) $m(ye) $m(xf) $m(yf)  \
           -fill $c3 -stipple "" -tag " A mm mxu$i ma$i "

        $m_canvas create line  $m(xc) $m(yc) $m(xd) $m(yd) \
           $m(xh) $m(yh) $m(xg) $m(yg)  \
           -fill $c3 -stipple "" -tag " A mm mxu$i mb$i "
        }


      set res 5; # the bad resolution 

     # central Bezier
     set x0 $m(x1); set y0 $m(y1)
     for {set j 0} {$j <= $res} {incr j 1} {
       set t [expr (1.0*$j) / (1.0*$res) ]
       set xt [expr (1-$t)*(1-$t)*$m(x1) + 2*(1-$t)*$t*$m(x2) + $t*$t*$m(x3) ] 
       set yt [expr (1-$t)*(1-$t)*$m(y1) + 2*(1-$t)*$t*$m(y2) + $t*$t*$m(y3) ] 
         $m_canvas create line $x0 $y0 $xt $yt \
                   -fill cyan  -width 1  -tag "A bc mxu$i"
         set x0 $xt; set y0 $yt
         }


     # blue Bezier
     set x0 $m(xb); set y0 $m(yb)
     for {set j 0} {$j <= $res} {incr j 1} {
       set t [expr (1.0*$j) / (1.0*$res) ]
       set xt [expr (1-$t)*(1-$t)*$m(xb) + 2*(1-$t)*$t*$m(xp) + $t*$t*$m(xf) ] 
       set yt [expr (1-$t)*(1-$t)*$m(yb) + 2*(1-$t)*$t*$m(yp) + $t*$t*$m(yf) ] 
         $m_canvas create line $x0 $y0 $xt $yt -fill $c3 \
                    -width 1 -tag "A bb mxu$i" 
         set x0 $xt; set y0 $yt
         }


     # red Bezier
     set x0 $m(xc); set y0 $m(yc)
     for {set j 0} {$j <= $res} {incr j 1} {
       set t [expr (1.0*$j) / (1.0*$res) ]
       set xt [expr (1-$t)*(1-$t)*$m(xc) + 2*(1-$t)*$t*$m(xq) + $t*$t*$m(xg) ] 
       set yt [expr (1-$t)*(1-$t)*$m(yc) + 2*(1-$t)*$t*$m(yq) + $t*$t*$m(yg) ] 
         $m_canvas create line $x0 $y0 $xt $yt -fill $c3 \
                   -width 1 -tag "A br mxu$i" 
         set x0 $xt; set y0 $yt
         }

        }

      }
    # puts stdout "plot mxu $i"
    }



     #-------------------------------------
     public method plot_label {oriA oriB} {
     #-------------------------------------
     # plots labels of current slit instance to canvas 

     # derive slit number from instance name
     set a [lindex [split $this : ] 6]  
     set i [string trimleft $a C_mxus]

     # highlight rim of the slit  
     switch -exact -- [$this cget -shape] {
       STRAIGHT {  

     if {[$this cget -ref_slit]} { set dummy $c4; set c4 cyan  }

                $m_canvas create poly $m(xb) $m(yb) $m(xc) $m(yc) \
                          $m(xg) $m(yg) $m(xf) $m(yf) -outline $c4 \
                          -fill ""  -tag "A mm mxu$i ms$i "

     # slit number label (for reference slit)
     # puts stdout "plot_mxu $i, [$this cget -ref_slit]"


     $m_canvas create text [expr ($m(xb)+$m(xf))/2] \
                          [expr ($m(yb)+$m(yf))/2] \
                          -anchor $oriA -fill $c4 -text " #$i " \
                          -tag "A rs mxu$i" 

     if {[$this cget -ref_slit]} { set c4 $dummy }
                }
           none     {  }
           CURVED   {  


     set res 200;  # the good resolution  


     # delete the low resolution curves 
     $m_canvas delete bc
     $m_canvas delete br
     $m_canvas delete bb

     # central Bezier
     set x0 $m(x1); set y0 $m(y1)
     for {set j 0} {$j <= $res} {incr j 1} {
       set t [expr (1.0*$j) / (1.0*$res) ]
       set xt [expr (1-$t)*(1-$t)*$m(x1) + 2*(1-$t)*$t*$m(x2) + $t*$t*$m(x3) ] 
       set yt [expr (1-$t)*(1-$t)*$m(y1) + 2*(1-$t)*$t*$m(y2) + $t*$t*$m(y3) ] 
     $m_canvas create line $x0 $y0 $xt $yt -fill cyan -width 1 -tag "A mm mxu$i" 
         set x0 $xt; set y0 $yt
         }

     # blue Bezier
     set x0 $m(xb); set y0 $m(yb)
     for {set j 0} {$j <= $res} {incr j 1} {
       set t [expr (1.0*$j) / (1.0*$res) ]
       set xt [expr (1-$t)*(1-$t)*$m(xb) + 2*(1-$t)*$t*$m(xp) + $t*$t*$m(xf) ] 
       set yt [expr (1-$t)*(1-$t)*$m(yb) + 2*(1-$t)*$t*$m(yp) + $t*$t*$m(yf) ] 
     $m_canvas create line $x0 $y0 $xt $yt -fill $c4 -width 1 -tag "A mm mxu$i" 
         set x0 $xt; set y0 $yt
         }

     # red Bezier
     set x0 $m(xc); set y0 $m(yc)
     for {set j 0} {$j <= $res} {incr j 1} {
       set t [expr (1.0*$j) / (1.0*$res) ]
       set xt [expr (1-$t)*(1-$t)*$m(xc) + 2*(1-$t)*$t*$m(xq) + $t*$t*$m(xg) ] 
       set yt [expr (1-$t)*(1-$t)*$m(yc) + 2*(1-$t)*$t*$m(yq) + $t*$t*$m(yg) ] 
     $m_canvas create line $x0 $y0 $xt $yt -fill $c4 -width 1 -tag "A mm mxu$i" 
         set x0 $xt; set y0 $yt
         }

         # top and bottom of curved slit 
         $m_canvas create line $m(xb) $m(yb) $m(xc) $m(yc) \
                   -fill $c4 -width 1 -tag "A mm mxu$i"
         $m_canvas create line $m(xf) $m(yf) $m(xg) $m(yg) \
                   -fill $c4 -width 1 -tag "A mm mxu$i"

     # slit number label 
     #$m_canvas create text [expr ($m(xb)+$m(xf))/2] \
     #                     [expr ($m(yb)+$m(yf))/2] \
     #                     -anchor $oriA -fill $c4 -text " #$i " \
     #                     -tag "A door mxu$i" 


     # slit number label (for reference slit)

     $m_canvas create text [expr ($m(xb)+$m(xf))/2] \
                          [expr ($m(yb)+$m(yf))/2] \
                          -anchor $oriA -fill $c4 -text " #$i " \
                          -tag "A rs mxu$i"

      plot_bez B00

        }
           CIRCLE   {  
                  # tcltk too stupid, cannot draw oval with angle
                  set lx [expr $m(xb)-$m(xg)]; set ly [expr $m(yb)-$m(yg)]
                  set lr [expr sqrt($lx*$lx+$ly*$ly)/2.0/sqrt(2.0)]
                  set lx [expr ($m(xb) +$m(xg)) / 2.0]
                  set ly [expr ($m(yb) +$m(yg)) / 2.0]

     # if {[$this cget -ref_slit]} { set dummy $c4; set c4 cyan  }
     
                  $m_canvas create oval [expr $lx-$lr] [expr $ly-$lr] \
                                        [expr $lx+$lr] [expr $ly+$lr] \
                     -outline $c4 -fill "" -tag " A mm mxu$i ms$i "

     # slit number label (for reference slit)
     # puts stdout "plot_mxu $i, [$this cget -ref_slit]"


     $m_canvas create text [expr ($m(xb)+$m(xf))/2] \
                          [expr ($m(yb)+$m(yf))/2] \
                          -anchor $oriA -fill $c4 -text " #$i " \
                          -tag "A rs mxu$i" 

     # if {[$this cget -ref_slit]} { set c4 $dummy }
                    }
            }



     # slit width label (to configure slit)
     $m_canvas create text [expr ($m(xc)+$m(xg))/2] \
                           [expr ($m(yc)+$m(yg))/2] \
                        -anchor $oriB -fill $c4 \
                        -text [format "%5.2f''"  [$this cget -slit_width]  ]\
                        -tag "A sw mxu$i"

     # east drive A label (used to de-activate the slit)
     $m_canvas create text [expr ($m(xa)+$m(xe))/2] \
                           [expr ($m(ya)+$m(ye))/2] \
                    -anchor $oriA -fill $c4 -text "#$i" \
                    -tag "A kill mxu$i"

     # west drive B label (used to de-activate the slit)
     $m_canvas create text [expr ($m(xd)+$m(xh))/2] \
                           [expr ($m(yd)+$m(yh))/2] \
                    -anchor $oriB  -fill $c4 -text "#$i" \
                    -tag "A kill mxu$i"

     if {$mmff} {
        # standard mask mode 
        # you are not allowed to modify a slit
        # you are not allowed to delete a slit
        $m_canvas dtag kill
        $m_canvas dtag sw 
        }

     }

     #------------------------
     public method plot_bez {tagname} {
     #------------------------
       # do not plot tagname
       # plot only the two other bezier points 
       # the graphical elements (one of the 
       # three cyan bezier points) is moved with 
       # the tk internal move command
       # but the shape of the curve will be updated
       # in mxu_class

      # convex hull 
      $m_canvas delete hull
      $m_canvas create line $m(x1) $m(y1) $m(x2) $m(y2) $m(x3) $m(y3) \
          -fill black -width 1 -tag "A mm mxu$slit_counter hull"

      set r [expr sqrt($sl_mm * $sw_mm) * 3  ]  

      # the three Bezier points 
      set a [string index $tagname 1] 
      # puts stdout "$tagname >$a<"
 
       if {$a != 1} {
       $m_canvas delete B13
       $m_canvas create oval [expr $m(x1)-$r] [expr $m(y1)-$r] \
                             [expr $m(x1)+$r] [expr $m(y1)+$r] \
                 -fill cyan -outline black -width 2 \
                 -tag "A door mxc$slit_counter B13" 
       }

       if {$a != 2} {
       $m_canvas delete B23
       $m_canvas create oval [expr $m(x2)-$r] [expr $m(y2)-$r] \
                             [expr $m(x2)+$r] [expr $m(y2)+$r] \
                 -fill cyan -outline black -width 2 \
                 -tag "A door mxc$slit_counter B23"
       }

       if {$a != 3} {
       $m_canvas delete B33
       $m_canvas create oval [expr $m(x3)-$r] [expr $m(y3)-$r] \
                             [expr $m(x3)+$r] [expr $m(y3)+$r] \
                 -fill cyan -outline black -width 2 \
                 -tag "A door mxc$slit_counter B33"
       }

      # the three labels for three Bezier points 
      # $m_canvas create text $m(x1) $m(y1) -anchor c -fill black 
      #                 -text "1" -tag "A mm mxu$slit_counter"
      # $m_canvas create text $m(x2) $m(y2) -anchor c -fill black 
      #                  -text "2" -tag "A mm mxu$slit_counter"
      # $m_canvas create text $m(x3) $m(y3) -anchor c -fill black 
      #                  -text "3" -tag "A mm mxu$slit_counter"

       }


# 4 ------------- PUT 

     #========================
     public method put_OC {a} {
     #========================
     # a = list with wcs coords 
     # corrected for optical image distortion
     # usually called after calc_zemax to store 
     # the values again 
     set oc_alp [lindex $a 0] 
     set oc_del [lindex $a 1] 
     # I know its a blame
     $m_image convert coords $oc_alp $oc_del "wcs $eqx" lx ly canvas
     set a [c2m $lx $ly] 
     # mmff OC corrected
     set oc_x [lindex $a 0]
     set oc_y [lindex $a 1]
     }

     #========================
     public method put_b_oc {nr val} {
     #========================
     # val is list of pixel coords coords 
     # just to store the two times (three bezier points) 
     # corrected for optical distortion
     $m_image convert coords [lindex $val 0] [lindex $val 1] image \
                                                       ca1 ca2 canvas 

     # the first three coord pairs for red  bezier curve
     # the last  three coord pairs for blue bezier curve

     switch -exact -- $nr {
     1 { set bf1 [c2m $ca1 $ca2] }
     2 { set bf2 [c2m $ca1 $ca2] }
     3 { set bf3 [c2m $ca1 $ca2] }
     4 { set bf4 [c2m $ca1 $ca2] }
     5 { set bf5 [c2m $ca1 $ca2] }
     6 { set bf6 [c2m $ca1 $ca2] }
       }

     }



# 5 --------------------- GET 
#   mmu coords of the MMU working table --> gerber file
#   OC  coords for the focal field      --> p_focf file 
#   normal                              -->   fims file 


# STRAIGHT and CIRCLE 

     #========================
     public method get_xy_mmu {} {
     #========================
     # returns slit center as list in mm of the MMU working table 
     # we use conversion X_mmu = Y_ff, 
     #                   Y_mmu = -X_ff
     return [list [format "%5.3f" $y] [format "%5.3f" [expr -1.0 * $x] ] ]
     }

     #========================
     public method get_xy {} {
     #========================
     # returns slit center as list in mm focplane units 
     return [list [format "%5.3f" $x] [format "%5.3f" $y] ]
     }

      
     #========================
     public method get_OC_RA {} {
     #========================
     # corrected for optical image distortion
     return $oc_alp }
      
     #========================
     public method get_OC_DEC {} {
     #========================
     # corrected for optical image distortion
     return $oc_del }

     #========================
     public method get_RA {} {
     #========================
	 
           return $alp
	
     }
      
     #========================
     public method get_DEC {} {
     #========================
	 
     return $del 
	 
     }

     #========================
     public method get_slit_mmu {} {
     #========================
     # slit position and full slit width/length  in mm MMU working table  
     return [list $y [expr -1.0 * $x] $sl_mm $sw_mm]  }

     #========================
     public method get_slit {} {
     #========================
     # slit position and full slit width in mm focal plane 
     return [list $x $y $sw_mm $sl_mm]  }

     #========================
     public method get_oc_slit_mmu {} {
     #========================
     # slit position and full slit width in mm MMU working table  
     # corrected for image distortion 
     return [list $oc_y [expr -1.0 * $oc_x]  $sl_mm $sw_mm]  }

     #========================
     public method get_oc_slit {} {
     #========================
     # slit position and full slit width in mm focal plane at position 
     # corrected for image distortion 
     return [list $oc_x $oc_y $sw_mm $sl_mm]  }


#  STRAIGHT POSANG != 0


     #========================
     public method get_oc_slit_contour_mmu {} {
     #========================
     # return the 4 corner points of rectangular slits 
     # at position corrected for optical image distortion 
     # ++++++++++++++++++++++++++++++
     # in mm of the MMU working table 
     # ++++++++++++++++++++++++++++++
     return [list [expr $s(yb)+$oc_y-$y] \
                 [expr -$s(xb)-$oc_x+$x] \
                  [expr $s(yc)+$oc_y-$y] \
                 [expr -$s(xc)-$oc_x+$x] \
                  [expr $s(yg)+$oc_y-$y] \
                 [expr -$s(xg)-$oc_x+$x] \
                  [expr $s(yf)+$oc_y-$y] \
                 [expr -$s(xf)-$oc_x+$x] ] }


     #========================
     public method get_slit_contour_mmu {} {
     #========================
     # return the 4 corner points of rectangular slits 
     # at position NOT corrected for optical image distortion 
     # when creating technical masks 
     # ++++++++++++++++++++++++++++++
     # in mm of the MMU working table 
     # ++++++++++++++++++++++++++++++
     return [list  $s(yb) \
                  [expr -1.0 * $s(xb)] \
                   $s(yc) \
                  [expr -1.0 * $s(xc)] \
                   $s(yg) \
                  [expr -1.0 * $s(xg)] \
                   $s(yf) \
                  [expr -1.0 * $s(xf)] ] }


     #========================
     public method get_oc_slit_contour {} {
     #========================
     # return the 4 corner points of rectangular slits 
     # at position corrected for optical image distortion 
     return [list [expr $s(xb)+$oc_x-$x] \
                  [expr $s(yb)+$oc_y-$y] \
                  [expr $s(xc)+$oc_x-$x] \
                  [expr $s(yc)+$oc_y-$y] \
                  [expr $s(xg)+$oc_x-$x] \
                  [expr $s(yg)+$oc_y-$y] \
                  [expr $s(xf)+$oc_x-$x] \
                  [expr $s(yf)+$oc_y-$y] ] }

     #========================
     public method get_slit_contour {} {
     #========================
     # return the 4 corner points of rectangular slits 
     # at position corrected for optical image distortion 
     return [list [expr $s(xb)-$x] \
                  [expr $s(yb)-$y] \
                  [expr $s(xc)-$x] \
                  [expr $s(yc)-$y] \
                  [expr $s(xg)-$x] \
                  [expr $s(yg)-$y] \
                  [expr $s(xf)-$x] \
                  [expr $s(yf)-$y] ] }


# CURVED 


     #========================
     public method get_curve_mm {} {
     #========================
     # return the three bezier points in focal field units
     lappend new $s(xb) $s(yb) $s(xp) $s(yp) $s(xf) $s(yf) 
     lappend new $s(xc) $s(yc) $s(xq) $s(yq) $s(xg) $s(yg)
     return $new
     }

     #========================
     public method get_curve_ca {} {
     #========================
     # return the six bezier points in canvas coords
     lappend new $m(xb) $m(yb) $m(xp) $m(yp) $m(xf) $m(yf) 
     lappend new $m(xc) $m(yc) $m(xq) $m(yq) $m(xg) $m(yg)
     return $new
     }


     #========================
     public method get_b_oc {} {
     #========================
     # return the six bezier points in ff focal plane, 
     # corrected for optical distortion
     # no longer in use 
     return [concat $bf1 $bf2 $bf3 $bf4 $bf5 $bf6 ]
     }

     #========================
     public method get_b_oc_mmu {} {
     #========================
     # return the six bezier points in ff focal plane, 
     # corrected for optical distortion
     # in units of the MMU working table 
     # y_mmu = x_ff
     # x_mmu = -y_ff
     lappend new [lindex $bf1 1] [expr -1.0 * [lindex $bf1 0] ] 
     lappend new [lindex $bf2 1] [expr -1.0 * [lindex $bf2 0] ] 
     lappend new [lindex $bf3 1] [expr -1.0 * [lindex $bf3 0] ] 
     lappend new [lindex $bf4 1] [expr -1.0 * [lindex $bf4 0] ] 
     lappend new [lindex $bf5 1] [expr -1.0 * [lindex $bf5 0] ] 
     lappend new [lindex $bf6 1] [expr -1.0 * [lindex $bf6 0] ] 
     return $new 
     }



# 7 --------------------  special Bezier


     #========================
     public method mod_bez { tagname shift_x shift_y } {
     #========================
     # modify shape of CURVED
     # used via interactive curser drags 
     # tagname = tag of bezier point,  must be B13 B23 or B33
     # shift_x shift_y = cursor shift in canvas
     
     # modify the curved slit online when cursor drag

     # convert a distance from canvas to mmfocf
     set dummy  [ c2m $shift_x $shift_y ]  
     set shaft_x [expr [lindex $dummy 0] - [lindex $mc_off 0] ]
     set shaft_y [expr [lindex $dummy 1] - [lindex $mc_off 1] ]

     switch -exact -- $tagname { 
      B13 { set x1 [expr $s(x1) + $shaft_x]
            set y1 [expr $s(y1) + $shaft_y]
          }
      B23 { set x2  [expr $s(x2) + $shaft_x]
            set y2  [expr $s(y2) + $shaft_y]
          }
      B33 { set x3 [expr $s(x3) + $shaft_x]
            set y3 [expr $s(y3) + $shaft_y]
          }
       }
     create_curve
     }

  #---------------------------------
  public method move_curve {curve} {
  #---------------------------------
         # modify shape of CURVED via new Bezier point 
         # positions. Used after CURVED has been created
         # used in load_mxu_TPX 

         if {[string match *CURVED* [$this cget -shape]]} {

         # configure a curved slit by points 
         set x1 [lindex $curve 0]
         set y1 [lindex $curve 1]
         set x2 [lindex $curve 2]
         set y2 [lindex $curve 3]
         set x3 [lindex $curve 4]
         set y3 [lindex $curve 5]
         
         create_curve
         }
         }




# 4a ------------special standard masks 

     public method M2C {} {
	 set dummy [m2c $x $y]
	 return $dummy
     }

     #============================
     public method M2W {{w_list ""}} {
     #===========================
     # convert the mmff coords (x,y) to ra,dec
     # call this before saving a standard mask 
     set dummy [m2c $x $y]

	 if {"$w_list" != ""} {	    	   
                 set alp [lindex $w_list 0]
                 set del [lindex $w_list 1]
	 } else {
	   # AAG PPRS 10127 
           if {[catch {$m_image convert coords \
			   [lindex $dummy 0] [lindex $dummy 1] canvas \
			   alp del "wcs $eqx"}]} {
               #TSZ modified text
	       info_dialog "$alp $del out of range!!! \n This reference slit will be ignored !! \n You may want to move the mask such that the reference slits are within the pre-image \n \n PRESS OK and then RESET MASK !" .skycat1
	       return 1
	   } else {
	       puts stdout "$alp $del OK"
	   }
       }
     return 0
 }





# 5 ------------miscellaneous 


     #============================
     public proc get_MASK_NAID {} {
     #===========================
     # return the ESO program ID and the MASK ID
     return $MASK_NAID
     }

     #============================
     public proc set_MASK_NAID {pm} {
     #===========================
     # set the MASK_NA the MASK_ID
     set MASK_NAID $pm
     # puts stdout ">$MASK_NAID<"
     }

     #============================
     public proc disp { 4val } {
     #===========================
     # read the grism dispersion [mm/nm], the grism lam0
     # the lam0 of the interference filter and 
     # the FWHM of the i-filter; all in nm 

     set disp [lindex $4val 0]
     set lcg  [lindex $4val 1]
     set lcf  [lindex $4val 2]
     set fwhm [lindex $4val 3]
     # puts stdout "mxu-class : $4val"
     }

     #============================
     public proc fc { focscale } {
     #===========================
     set fosc $focscale
     }

     #====================================
     public proc coeff { mx1 my1 cx1 cy1 \
                     mx2 my2 cx2 cy2 \
                     mx3 my3 cx3 cy3  } {
     #=====================================
     coeff_mc $mx1 $my1 $cx1 $cy1 \
              $mx2 $my2 $cx2 $cy2 \
              $mx3 $my3 $cx3 $cy3

     set mc_off [ c2m 0.0 0.0 ]
     }

     #===================
     public proc get_mmff {} {
     #===================
     # change status od mxu 
     return $mmff
     }

     #===================
     public proc set_mmff {flag} {
     #===================
     # change status of mxu 
     set mmff $flag
     }


     #===================
     public proc label {flag} {
     #===================
     # change label for all instances 
     set label $flag
     }


     #===================
     public proc put_col {col} {
     #===================
     # sets color for all mxu slits
     set c1 [lindex $col 0]
     set c2 [lindex $col 1]
     set c3 [lindex $col 2]
     set c4 [lindex $col 3]
     set trapa [lindex $col 4]
     }


     #===================
     public proc nos {} {
     #===================
     # returns the number of slits 
     # the number of all instances, 
     # wether they have been deleted in the meanwhile (ON==0)
     # or if they are still available (ON==1)
     return $slit_counter
     }

     #===================
     public proc nor {} {
     #===================
     return $ref_slit_cou
     }


     #===================
     public method erase {} {
     #===================
     # erase individual slit, circle or curve  from canvas 

     # derive slit number from instance name 
     set a [lindex [split $this : ] 6]  
     set i [string trimleft $a C_mxus]

     # delete from canvas 
     $m_canvas delete mxu$i
     if {[string match *CURVED* [$this cget -shape]]} {
        $m_canvas delete mxc$i
        }
     }


     #========================
     public method get_add_mmu {} {
     #========================
     # puts stdout "$slit_counter [$this cget -ON] $add_mmu"
     return $add_mmu  
     }

     #========================
     public method get_add {} {
     #========================
     # puts stdout "$slit_counter [$this cget -ON] $add"
     return $add  
     }

     #========================
     public method test {} {
     #========================
     puts stdout "Hello World: You have called method test" 
     puts stdout "$slit_number instances"
     puts stdout "$p1"
     }

     #================================
     public method status {} {
     #================================
     # maintenance 

     # derive slit number from instance name
     set a [lindex [split $this : ] 6]  
     set i [string trimleft $a C_mxus]

     puts stdout [format "%2d/%2d %s" $i $slit_counter [$this cget -shape] ]
     
     switch -exact -- [$this cget -shape] {
     STRAIGHT {
#          puts stdout [format "f=%7.3f,%7.3f co=%7.3f,%7.3f" \
#                          $x $y $oc_x $oc_y ]
           puts stdout [format "f=%7.3f,%7.3f" $x $y  ]
           puts stdout [format "W=%7.3f''=%7.3fmm L=%7.3f''=%7.3fmm P=%7.3fo" \
                 [$this cget -slit_width] $sw_mm  \
                 [$this cget -slit_length] $sl_mm \
                 [$this cget -pos_ang] ]
              }
     CIRCLE   {
#          puts stdout [format "f=%7.3f,%7.3f co=%7.3f,%7.3f" \
#                          $x $y $oc_x $oc_y ]
           puts stdout [format "f=%7.3f,%7.3f" $x $y  ]
           puts stdout [format "W=%7.3f''=%7.3fmm"  \
                 [$this cget -slit_width] $sw_mm  ]
              }
     CURVED   {
           puts stdout [format "f=%7.3f,%7.3f co=%7.3f,%7.3f" \
                           $s(xb) $s(yb) [lindex $bf1 0] [lindex $bf1 1] ]
           puts stdout [format "f=%7.3f,%7.3f co=%7.3f,%7.3f" \
                           $s(xp) $s(yp) [lindex $bf2 0] [lindex $bf2 1] ]
           puts stdout [format "f=%7.3f,%7.3f co=%7.3f,%7.3f" \
                           $s(xf) $s(yf) [lindex $bf3 0] [lindex $bf3 1] ]
           puts stdout [format "f=%7.3f,%7.3f co=%7.3f,%7.3f" \
                           $s(xc) $s(yc) [lindex $bf4 0] [lindex $bf4 1] ]
           puts stdout [format "f=%7.3f,%7.3f co=%7.3f,%7.3f" \
                           $s(xq) $s(yq) [lindex $bf5 0] [lindex $bf5 1] ]
           puts stdout [format "f=%7.3f,%7.3f co=%7.3f,%7.3f" \
                           $s(xg) $s(yg) [lindex $bf6 0] [lindex $bf6 1] ]

           puts stdout [format "W=%7.3f''=%7.3fmm"  \
                 [$this cget -slit_width] $sw_mm  ]
              }
     }
 
     }


  #++++++++++++++++++++++++++
  # configuration options 
  #++++++++++++++++++++++++++

  public  variable eqx 2000.0
  
  #----------------------------------
  public variable slit_length 10.0 {
  #----------------------------------
          # from arcsec to mmff

          # slit length has only a meaning for STRAIGHT
          # for CIRCLE and CURVED, slit_length has no meaning 

          if {[string match *STRAIGHT* [$this cget -shape]]} {
          if {$mmff} {
             # the slit length is given in mm
             set sl_mm $slit_length
             set slit_length [ expr $sl_mm * $fosc ]
             } else {
             set sl_mm [ expr $slit_length / $fosc ]
             }

            
          if {$slit_initialized} {
             create_slit 
             # create slit does not create a new instance 
             # it does only update the  mm-focplane (s-array)
             # and converts from mm focplane (s-array) 
             # to canvas (m-array)
             }

             } else {

             # current slit shape is CIRCLE or CURVED
             # we assume only circle 
             set slit_length [$this cget -slit_width]
             set sl_mm $sw_mm 
             }

           # puts stdout "sl:$slit_length"
           }

  
  #----------------------------------
  public variable slit_width 1.0 {
  #----------------------------------
          # from arcsec to mmff
          if {$mmff} {
             # slit width is given in mm 
             set sw_mm $slit_width
             set slit_width [ expr $sw_mm * $fosc ]
             } else {
             set sw_mm [ expr $slit_width / $fosc ]
             }

          if {$slit_initialized} {
             # only for later config, not during slit create
             switch -exact -- [$this cget -shape] {
              STRAIGHT { create_slit }
              none     { }
              CURVED   { create_curve }
              CIRCLE   { set sl_mm $sw_mm
                         set slit_length  $slit_width
                         create_circle }
                 }
             }
           # puts stdout "sw:$slit_width"
           }

  public variable ff   

  #----------------------------
  public variable pos_ang 0.0 {
  #----------------------------
         # from deg to rad 
         set pa [expr $pos_ang * 0.0174532]
          if {$slit_initialized} {
             create_slit
             }
           # puts stdout "pa:$pos_ang"
         }



  public variable child;           # to whom I give my gens 
  public variable shape STRAIGHT;  # slit shape: STRAIGHT or CURVED or CIRCLE
  public variable ref_slit 0 {

         if {$slit_initialized} {
            # not called by constructor 
            if { [$this cget -ref_slit] } {
               # ref_slit activated 
               incr ref_slit_cou 1;
               } else {
               # ref_slit de-selected
               incr ref_slit_cou -1;
               }
            }
             
            }


  public variable ON 1 {
         # puts stdout "ON:$ON"
         # you cannot delete a slit
         # you can only switch it off or on
         # because of renumbering problems 
         }

  public variable target_name NN  {
         # puts stdout "tn:$target_name"
         }


  private variable alp;               # RA of target in slit, in wcs  
  private variable del;               # DEC of target in slit, in wcs 
                                      # for CURVED : RA,DEC for t=0.5
  private variable oc_alp;            # RA of target optical image 
                                      # distortion correction applied
  private variable oc_del;            # DEC of target, with oc  
  private variable x;                 # slit central position in mm_ff
  private variable y;                 # slit central position in mm_ff 
  private variable oc_x;              # central position in mm_ff, OC applied 
  private variable oc_y;              # central position in mm_ff, OC applied 
  private variable x1;                # First gridpoint for bezier curved slit 
  private variable y1;                # First gridpoint for bezier curved slit
  private variable x2;                # Middle gridpoint for bezier curved slit 
  private variable y2;                # Middle gridpoint for bezier curved slit
  private variable x3;                # Last gridpoint for bezier curved slit
  private variable y3;                # Last gridpoint for bezier curved slit
  private variable s;                 # slit dimension array in mm_ff units 
  private variable m;                 # slit dimension array in canvas units 
  private variable m_canvas;          # to plot 
  private variable m_image;           # to convert 
  private variable slit_initialized 0;# flag 
  private variable sw_mm;             # full slit width  in mm focal plane
  private variable sl_mm;             # full slit length in mm focal plane
  private variable pa;                # slit position angle in rad   
  private variable res 30;            # Bezier resolution 
  private variable add "";            # aperture definition 
  private variable add_mmu "";        # aperture definition 
  private variable mc_off;            # offset for convert distance 
  private variable bf1
  private variable bf2
  private variable bf3
  private variable bf4
  private variable bf5
  private variable bf6

  private common MASK_NAID {M 999999}

  private common slit_counter 0;      # number of slits 
  private common ref_slit_cou 0;      # number of ref_slits 
  private common fosc;                # focal scale

  # used when MXU is used with interference filters
  private common disp;                # dispersion in mm/A
  private common lcg;                 # lambdacenter of grism                
  private common lcf;                 # lambdacenter of interference filter
  private common fwhm;                # fwhm of interference filter

# graphical layout common for all slits 
  private common c1 red; 
  private common c2 yellow; 
  private common c3 red; 
  private common c4 yellow; 
  private common trapa gray25; 

  private common mmff 0;         # input x,y =mmff or canvas 
  private common label 0;        # graphical label            


     }
     # end of the class MXU_SLIT

}
# end of namespace

#
# ___oOo___

